/*
 * Copyright (C) 1997-2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include "config.h"
#include <readline/readline.h>
#include <readline/history.h>
#include "client.cmdline.h"
#include "crypt.h"


/* A static variable for holding the line. */
static char *line_read = (char *)NULL;

struct gengetopt_args_info args_info;

/*
 * client log function
 */
void para_log(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;

	/* ignore log message if loglevel is not high enough */
	if (ll < args_info.loglevel_arg)
		return;
	if (ll < WARNING)
		outfd = stdout;
	else
		outfd = stderr;
	va_start(argp, fmt);
	vfprintf(stdout, fmt, argp);
	va_end(argp);
}

/*
 * Read a string, and return a pointer to it. Returns NULL on EOF.
 */
char *rl_gets(void)
{
	/* If the buffer has already been allocated, return the memory
	 * to the free pool.
	 */
	if (line_read) {
		free (line_read);
		line_read = (char *)NULL;
	}
	/* Get a line from the user. */
	line_read = readline ("para_client> ");
	/* If the line has any text in it, save it on the history. */
	if (line_read && *line_read)
		add_history (line_read);
	return line_read;
}

/*
 * do several cleanups on sigint
 */
void signals_int(int i)
{
	rl_save_prompt();
	printf("\n");
	rl_message("\n");
	rl_kill_full_line(0, 0);
	rl_restore_prompt();
	rl_free_line_state();
	rl_cleanup_after_signal();
	rl_reset_after_signal();
	//printf("Caught SIGINT\n");
}

/*
 * return proper exit status
 */
int get_exit_status (int status)
{
	if (WIFEXITED(status))
		return WEXITSTATUS(status);
	if (WIFSIGNALED(status))
		return WTERMSIG(status);
	return WSTOPSIG(status);
}

/*
 * wait and tell exit status
 */
void sigchld_handler(int s)
{
	int i, es;
	pid_t pid;

	pid = wait(&i);
	es = get_exit_status(i);
	//printf("verbose=%i\n",verbose);
	PARA_DEBUG_LOG("child process %i died. Exit status: %i (%s)\n",
		pid, es, strerror(es));
}

void get_options(int argc, char *argv[],
	char **config_file, char **key_file)
{
	char *home;
	static char default_key_file[_POSIX_PATH_MAX] = "";
	static char default_config_file[_POSIX_PATH_MAX] = "";
	struct stat statbuf;
	int ret;

	cmdline_parser(argc, argv, &args_info);
	if (!args_info.user_given)
		args_info.user_arg = para_logname();
	if (!args_info.key_file_given) {
		home = para_homedir();
		sprintf(default_key_file, "%s/.paraslash/key.%s", home,
			args_info.user_arg);
		free(home);
	}
	if (!args_info.config_file_given) {
		home = para_homedir();
		sprintf(default_config_file, "%s/.paraslash/client.conf",
			home);
		free(home);
	}
	if (!args_info.config_file_given)
		*config_file = default_config_file;
	else
		*config_file = args_info.config_file_arg;
	ret = stat(*config_file, &statbuf);
	if (ret && args_info.config_file_given) {
		fprintf(stderr, "can not stat config file %s\n",
			args_info.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (!ret)
		cmdline_parser_configfile(*config_file, &args_info, 0, 0, 0);
	if (!args_info.key_file_given)
		*key_file = default_key_file;
	else
		*key_file = args_info.key_file_arg;
}


/*
 * MAIN
 */
int main(int argc, char *argv[])
{

	struct sigaction sa, sa_chld;
	int sockfd, numbytes, i, interactive, received;
	struct hostent *he;
	struct sockaddr_in their_addr;
	char *command = NULL;
	char buf[8192];
	char challenge_decrypt[STRINGSIZE];
	char auth_str[MAXLINE]; /* FIXME: use send_va_buffer instead */
	char *key_file, *config_file;
	long unsigned challenge_nr;
	char *line;


	get_options(argc, argv, &config_file, &key_file);
	if (args_info.loglevel_arg <= NOTICE)
		cmdline_parser_print_version();
	PARA_INFO_LOG(
		"current loglevel: %d\n"
		"using config_file: %s\n"
		"using key_file: %s\n"
		"connecting to %s:%d\n",
		args_info.loglevel_arg,
		config_file,
		key_file,
		args_info.hostname_arg,
		args_info.server_port_arg
	);


	//setlinebuf(stdout);
	/* compute auth string to be sent to server */
	sprintf(auth_str, "auth %s", args_info.user_arg);
	/* install sigchld handler */
	sa_chld.sa_handler = sigchld_handler;
	sigemptyset(&sa_chld.sa_mask);
	sa_chld.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa_chld, NULL) == -1) {
		PARA_EMERG_LOG("%s", "failed to install sigchld handler\n");
		exit(EXIT_FAILURE);
	}
	interactive = args_info.inputs_num == 0? 1 : 0;
	if (interactive) {
		/* catch SIGINT only if in interactive mode */
		sa.sa_flags = 0;
		sa.sa_handler = signals_int;
		sigaction(SIGINT, &sa, NULL);
		PARA_NOTICE_LOG("%s", "no command, entering interactive mode\n");
	} else {
		/* not interactive, copy command and args to command */
		for (i = 0; i < args_info.inputs_num; i++) {
			size_t len = strlen(args_info.inputs[i]);
			if (command) {
				command = realloc(command, strlen(command)
					+ len + 2);
				strcat(command, "\n");
			} else {
				command = malloc((len + 2) * sizeof(char));
				*command = '\0';
			}
			strcat(command, args_info.inputs[i]);
		}
		//PARA_NOTICE_LOG("%s: Using command: %s\n", __func__, command);
	}
	//exit(1);
interactive_loop:
	if (interactive) { /* FIXME: That is crappy */
		int i = 0;
		char *p;

		/* read a line via readline */
		line = rl_gets();
		if (!line)
			return 0;
		if (!line[0])
			goto interactive_loop;
		p = line;
		while (sscanf(p, "%200s%n", buf, &i) == 1) {
			size_t len = strlen(buf);
			if (command) {
				command = realloc(command, strlen(command)
					+ len + 2);
				strcat(command, "\n");
			} else {
				command = malloc((len + 2) * sizeof(char));
				*command = '\0';
			}
			strcat(command, buf);
			p += i;
		}
//		printf("%s\n", command);
	}
	/* get the host info */
	PARA_NOTICE_LOG("getting host info of %s\n",
		args_info.hostname_arg);
	if (!(he = get_host_info(args_info.hostname_arg)))
		exit(EXIT_FAILURE);
	/* get new socket */
	if ((sockfd = get_socket()) < 0)
		exit(EXIT_FAILURE);
	/* init their_addr */
	init_sockaddr(&their_addr, args_info.server_port_arg, he);
	/* Connect */
	PARA_NOTICE_LOG("connecting to %s...\n",
		args_info.hostname_arg);
	if (do_connect(sockfd, &their_addr) < 0)
		exit(EXIT_FAILURE);
	/* Receive Welcome message */
	if ((numbytes = recv_buffer(sockfd, buf, sizeof(buf))) < 0)
		exit(EXIT_FAILURE);
	/* send auth command */
	PARA_INFO_LOG("<-- %s--> %s\n", buf, auth_str);
	if (send_buffer(sockfd, auth_str) < 0)
		exit(EXIT_FAILURE);
	/* receive challenge number */
	if ((numbytes = recv_buffer(sockfd, buf, sizeof(buf))) < 0)
		exit(EXIT_FAILURE);
	if (numbytes != 64) {
		PARA_EMERG_LOG("did not receive valid challenge (got %i bytes)\n",
			numbytes);
		buf[numbytes] = '\0';
		PARA_ERROR_LOG("received the following instead: %s\n", buf);
		goto write_out;
	}
	PARA_INFO_LOG("<-- [challenge (%i bytes)]\n", numbytes);
	/* decrypt challenge number */
	if (para_decrypt(key_file, &challenge_nr, (unsigned char *) buf, numbytes) < 0) {
		PARA_EMERG_LOG("%s", "decrypt error. Bad secret key?\n");
		exit(EXIT_FAILURE);
	}
	/* send decrypted challenge */
	sprintf(challenge_decrypt, "challenge_response:%lu", challenge_nr);
	PARA_INFO_LOG("--> %lu\n", challenge_nr);
	if (send_buffer(sockfd, challenge_decrypt) < 0)
		exit(EXIT_FAILURE);
	/* Wait for approval */
	PARA_NOTICE_LOG("%s", "waiting for approval from server\n");
	if ((numbytes = recv_buffer(sockfd, buf, sizeof(buf))) < 0)
		exit(EXIT_FAILURE);
	PARA_INFO_LOG("++++ server info ++++\n%s\n++++ end of server "
		"info ++++\n", buf);
	/* Check if server has sent "Proceed" message */
	if (!strstr(buf, PROCEED_MSG)) {
		PARA_EMERG_LOG("%s", "authentication failed\n");
		exit(EXIT_FAILURE);
	}
	/* send command */
	PARA_INFO_LOG("--> %s\n", command);
	if (send_buffer(sockfd, command) < 0)
		exit(EXIT_FAILURE);
	free(command);
	command = NULL;
	if (send_buffer(sockfd, EOC_MSG "\n") < 0)
		exit(EXIT_FAILURE);
	PARA_NOTICE_LOG("%s", "command sent.\n");
write_out:
	received = 0;

	PARA_INFO_LOG("bufsize: %d\n", sizeof(buf));
	/* write server output to stdout */
	while ((numbytes = recv_bin_buffer(sockfd, buf, sizeof(buf))) > 0) {
		int ret;

		if (!received && strstr(buf, AWAITING_DATA_MSG)) {
			PARA_NOTICE_LOG("%s", "<-- awaiting data\n");
			PARA_NOTICE_LOG("%s", "--> sending stdin\n");
			while ((ret = read(STDIN_FILENO, buf,
					sizeof(buf))) > 0)
				send_bin_buffer(sockfd, buf, ret);
			PARA_NOTICE_LOG("%s", "closing connection\n");
			numbytes = 1;
			break;
		}
		received = 1;
		if (write(STDOUT_FILENO, buf, numbytes) != numbytes)
			exit(EXIT_FAILURE);
	//	sleep(2);
	}
	if (numbytes < 0)
		exit(EXIT_FAILURE);
	if (!numbytes)
		PARA_NOTICE_LOG("%s", "connection closed by peer\n");
	close(sockfd);
	if (interactive)
		goto interactive_loop;
	return 0;
}
